local LCA = LibCombatAlerts

CombatAlerts2 = {
	ID = "CombatAlerts2",
	NAME = "CombatAlerts",
	EXPECTED_VERSION = 118090,
	URL = "https://www.esoui.com/downloads/info1855.html",

	activeModules = { },
	registeredModules = { },
	registeredZones = { },
	devModule = nil,

	isLegacyActive = false,
}
local CA1 = CombatAlerts
local CA2 = CombatAlerts2


--------------------------------------------------------------------------------
-- Bootstrap, modules, and legacy mode
--------------------------------------------------------------------------------

EVENT_MANAGER:RegisterForEvent(CA2.ID, EVENT_ADD_ON_LOADED, function( eventCode, addonName )
	if (addonName ~= CA2.NAME) then return end

	EVENT_MANAGER:UnregisterForEvent(CA2.ID, EVENT_ADD_ON_LOADED)

	-- Make sure Minion didn't mess up the manifest
	if (LCA.GetAddOnVersion(CA2.NAME) ~= CA2.EXPECTED_VERSION) then
		LCA.RunAfterInitialLoadscreen(function()
			CA2.ChatMessage(GetString(SI_CA_CORRUPTED), true)
		end)
		return
	end

	-- Legacy stuff
	CA2.HandleLegacyAddOns()
	CA1.OnAddOnLoaded()

	-- Initialize settings
	CA2.LoadSettings()
	LCA.RunAfterInitialLoadscreen(CA2.RegisterSettingsPanel)

	LCA.MonitorZoneChanges(CA2.ID, CA2.OnZoneChange)
end)

function CA2.OnZoneChange( zoneId )
	-- Legacy bootstrap
	CA1.Initialize()

	if (CA2.activeModules[1]) then
		for _, module in ipairs(CA2.activeModules) do
			module:Unload()
		end
		CA2.CleanupControls()
		CA2.DisablePurgeTracker()
	end

	local modules = CA2.registeredZones[zoneId]
	if (modules) then
		CA2.ToggleLegacy(false)
		for _, module in ipairs(modules) do
			module:Load()
		end
		CA2.activeModules = modules
	else
		CA2.activeModules = { }
		CA2.ToggleLegacy(true)
	end
end

function CA2.RegisterModule( template )
	local module = template:New()
	if (module.ID == "CA_M_DEV") then
		CA2.devModule = module
	elseif (not CA2.registeredModules[module.ID]) then
		CA2.registeredModules[module.ID] = module
		for _, zoneId in ipairs(module.ZONES) do
			if (CA2.registeredZones[zoneId]) then
				table.insert(CA2.registeredZones[zoneId], module)
			else
				CA2.registeredZones[zoneId] = { module }
			end
		end
	end
end

function CA2.FriendlyModuleName( update, zoneId, achievementId )
	if (GetAPIVersion() >= 101000 + update) then
		return string.format("U%d: %s", update, zoneId and LCA.GetZoneName(zoneId) or GetAchievementCategoryInfo(GetCategoryInfoFromAchievementId(achievementId)))
	else
		return string.format("U%d", update)
	end
end

function CA2.ToggleLegacy( enable )
	if (enable and not CA2.isLegacyActive) then
		CA2.isLegacyActive = true
		EVENT_MANAGER:RegisterForEvent(CA1.name, EVENT_PLAYER_ACTIVATED, CA1.PlayerActivated)
		EVENT_MANAGER:RegisterForEvent(CA1.name, EVENT_PLAYER_COMBAT_STATE, CA1.PlayerCombatState)
		CA1.PlayerActivated()
	elseif (not enable and CA2.isLegacyActive) then
		CA2.isLegacyActive = false
		EVENT_MANAGER:UnregisterForEvent(CA1.name, EVENT_PLAYER_ACTIVATED)
		EVENT_MANAGER:UnregisterForEvent(CA1.name, EVENT_PLAYER_COMBAT_STATE)
	end
end


--------------------------------------------------------------------------------
-- Chat messages
--------------------------------------------------------------------------------

function CA2.ChatMessage( message, useTag )
	CHAT_ROUTER:AddSystemMessage(string.format("[%s] %s", useTag and GetString(SI_CA_TITLE) or os.date("%H:%M:%S", GetTimeStamp()), message))
end


--------------------------------------------------------------------------------
-- Utilities
--------------------------------------------------------------------------------

function CA2.FindClosestGroupMember( )
	local resultDistance = -1
	local resultUnitTag = ""

	for i = 1, GetGroupSize() do
		local unitTag = GetGroupUnitTagByIndex(i)
		if (not AreUnitsEqual(unitTag, "player")) then
			local distance = LCA.GetDistance("player", unitTag, false, true)
			if (distance >= 0 and (distance < resultDistance or resultDistance == -1)) then
				resultDistance = distance
				resultUnitTag = unitTag
			end
		end
	end

	return resultDistance, resultUnitTag
end


--------------------------------------------------------------------------------
-- Handle conflicts with legacy addons
--------------------------------------------------------------------------------

function CA2.HandleLegacyAddOns( )
	if (type(PurgeTrackerData) == "table" and PurgeTrackerData.zones) then
		PurgeTrackerData.zones[725] = nil
	end
end


--------------------------------------------------------------------------------
-- Developer functions
--------------------------------------------------------------------------------

function CA2.DevModule( enable )
	if (CA2.devModule) then
		if (enable) then
			CA2.devModule:Load()
		else
			CA2.devModule:Unload()
		end
	end
end

function CA2.DevGetDebug( )
	local activeModule = CA2.activeModules[1]
	if (activeModule) then
		return activeModule, activeModule:GetDebugData()
	end
end
