local LCA = LibCombatAlerts
local CA1 = CombatAlerts
local CA2 = CombatAlerts2


--------------------------------------------------------------------------------
-- Defaults
--------------------------------------------------------------------------------

local DEFAULTS = {
	suppressModuleMessages = false,
}


--------------------------------------------------------------------------------
-- Initialization
--------------------------------------------------------------------------------

function CA2.LoadSettings( )
	CA2.sv = ZO_SavedVars:NewAccountWide("CombatAlertsSettings", 1, nil, DEFAULTS, nil, "$InstallationWide")

	-- Import settings from legacy, if they differ from their defaults
	if (not CA2.sv.statusPanel and (CA1.vars.panelTop ~= CA1.defaults.panelTop or CA1.vars.panelLeft ~= CA1.defaults.panelLeft)) then
		CA2.sv.statusPanel = { top = CA1.vars.panelTop, left = CA1.vars.panelLeft }
	end
	--[[
	if (not CA2.sv.nearby) then
		CA2.sv.nearby = { enabled = CA1.vars.nearbyEnabled }
		if (CA1.vars.nearbyTop ~= CA1.defaults.nearbyTop or CA1.vars.nearbyLeft ~= CA1.defaults.nearbyLeft) then
			CA2.sv.nearby.top = CA1.vars.nearbyTop
			CA2.sv.nearby.left = CA1.vars.nearbyLeft
		end
	end
	--]]
end


--------------------------------------------------------------------------------
-- Settings Panel
--------------------------------------------------------------------------------

function CA2.RegisterSettingsPanel( )
	local LAM = LibAddonMenu2

	if (LAM) then
		local panelId = "CombatAlertsSettingsPanel"

		CA2.settingsPanel = LAM:RegisterAddonPanel(panelId, {
			type = "panel",
			name = GetString(SI_CA_TITLE),
			version = LCA.FormatVersion(LCA.GetAddOnVersion(CA2.NAME)),
			author = "@code65536",
			website = CA2.URL,
			donation = CA2.URL .. "#donate",
			registerForRefresh = true,
		})

		LAM:RegisterOptionControls(panelId, {
			--------------------------------------------------------------------
			{
				type = "header",
				name = SI_GAME_MENU_SETTINGS,
			},
			--------------------
			{
				type = "checkbox",
				name = SI_CA_SETTINGS_SUPPRESS,
				getFunc = function() return CA2.sv.suppressModuleMessages end,
				setFunc = function(enabled) CA2.sv.suppressModuleMessages = enabled end,
			},

			--------------------------------------------------------------------
			{
				type = "header",
				name = SI_CA_SETTINGS_MODULES,
			},
			--------------------
			unpack(CA2.BuildModulesSettings())
		})
	end
end

function CA2.BuildModulesSettings( )
	local controls = { }

	for _, moduleId in ipairs(LCA.GetSortedKeys(CA2.registeredModules)) do
		local module = CA2.registeredModules[moduleId]

		local zoneNames = { }
		for _, zoneId in ipairs(module.ZONES) do
			table.insert(zoneNames, LCA.GetZoneName(zoneId, true))
		end

		local settingsControls = module:GetSettingsControls()
		if (#settingsControls > 0) then
			table.insert(settingsControls, 1, { type = "header", name = SI_GAME_MENU_SETTINGS })
		end

		table.insert(controls, {
			type = "submenu",
			name = CA2.registeredModules[moduleId].NAME,
			controls = {
				----------------------------------------------------------------
				{
					type = "description",
					text = zo_strformat(SI_CA_SETTINGS_MODULE_INFO, table.concat(zoneNames, ", "), module.AUTHOR)
				},
				--------------------
				{
					type = "checkbox",
					name = SI_ADDON_MANAGER_ENABLED,
					getFunc = function() return true end,
					setFunc = function(enabled) end,
					disabled = function() return true end,
				},
				--------------------
				unpack(settingsControls)
			},
		})
	end

	return controls
end
