local LCA = LibCombatAlerts
local CA1 = CombatAlerts
local CA2 = CombatAlerts2
local Module = CA_Module:Subclass()

Module.ID = "CA_M_U42"
Module.NAME = CA2.FriendlyModuleName(42, 1478)
Module.AUTHOR = "@code65536"
Module.ZONES = {
	1478, -- Lucent Citadel
}

Module.MONITOR_UNIT_IDS = true

Module.TIMER_ALERTS = {
	[217971] = { -2, 0, false, { 1, 0, 0.6, 0.8 } }, -- Heavy Strike
	[218274] = { -2, 2 }, -- Shear (Count Ryelaz)
	[222605] = { -2, 2 }, -- Shear (Baron Rize)
	[218672] = { -2, 2 }, -- Uppercut
	[218967] = { -2, 2 }, -- Uppercut
	[219030] = { -2, 2 }, -- Power Bash
	[219139] = { -2, 2 }, -- Smite
	[219420] = { 0, 0, false, { 1, 0, 0.6, 0.8 } }, -- Smite
	[219793] = { -2, 2 }, -- Crushing Shards
	[221863] = { -2, 0, false, { 1, 0, 0.6, 0.8 } }, -- Heavy Attack
	[221881] = { -2, 0, false, { 1, 0, 0.6, 0.8 } }, -- Frenzy
	[222271] = { 0, 0, false, { 1, 0, 0.6, 0.8 } }, -- Heavy Strike
	[222559] = { -2, 0, false, { 1, 0, 0.6, 0.8 } }, -- Batter
}

Module.DEFAULT_SETTINGS = {
	extraAlerts = true,
	cooldownPanel = true,
	passageProtection = false,
}

local STRINGS = {
	-- Extracted
	["8290981-0-121484"] = { default = "Shardborn Lightweaver^f", de = "Scherbensaat-Lichtweberin^f", es = "tejedora de luz nacida de la esquirla^f", fr = "tisse-lumière Enfant du fragment^f", jp = "シャードボーン・ライトウィーバー^f", ru = "Ткачиха света из клана Детей Осколков^f", zh = "碎晶族织光者" },
	["8290981-0-121485"] = { default = "Gloomy Blackguard^f", de = "düstere Schwarzwache^f", es = "guardia negra lúgubre^f", fr = "garde-obscurité sinistre^m", jp = "薄暗いブラックガード^f", ru = "Сумрачный страж мрака^f", zh = "幽暗黑卫" },
	["74865733-0-34170"] = { default = "Claim", de = "Einlösen", es = "Reclamar", fr = "Récupérer", jp = "獲得", ru = "Взять", zh = "获取" },
	["87370069-0-34170"] = { default = "Arcane Knot", de = "arkaner Knoten^m", es = "nudo arcano^m", fr = "nœud arcanique^md", jp = "アルケインの結び目", ru = "Тайный узел", zh = "奥术结" },

	-- Custom
	noCarrier = { default = "NO CARRIER" },
	nextFlux = { default = "to next spawn" },
	nextTempest = { default = "to next lines" },
	cooldownLabel = { default = "Cooldown: %s" },
	extraAlerts = { default = "Enable additional alerts" },
	cooldownPanel = { default = "Enable group-wide cooldown panel" },
	passageProtection = { default = "Enable Splintered Passage protection" },
}

local COLOR_L32 = 0xFFFF66FF
local COLOR_L24 = 0xFFFF66
local COLOR_D32 = 0xCC3333FF
local COLOR_D24 = 0xCC3333

local Data = {
	-- General
	extraAlertsTooltipIds = { 214237, 214254, 218283, 221945, 219179 },
	refresh = 222903,
	darkness = 214338,

	-- Boss 1
	boss1Abilities = {
		-- Abilities unique to the first boss for identification purposes
		[222277] = true, -- Twin Jab
		[222278] = true, -- Sting
	},
	mirrorSide = {
		light = 219329,
		dark = 219330,
	},
	annihilation = {
		[214187] = "dark", -- Brilliant Annihilation
		[214203] = "light", -- Bleak Annihilation
	},
	luster = {
		[214237] = COLOR_L32, -- Brilliant Lusterbeam
		[214254] = COLOR_D32, -- Bleak Lusterbeam
	},
	summons = {
		[218109] = {
			name = LCA.GetLocalizedData(STRINGS["8290981-0-121485"]),
			side = "light",
			color = COLOR_L32,
		},
		[218113] = {
			name = LCA.GetLocalizedData(STRINGS["8290981-0-121484"]),
			side = "dark",
			color = COLOR_D32,
		},
	},
	shear = {
		[218274] = true,
		[222605] = true,
	},
	meteor = 218283,
	boss1Colors = { COLOR_L24, COLOR_D24 },

	-- Boss 2
	fateSealer = 214136,
	jump = {
		id = 214383,
		escape = {
			[217987] = true, -- Defeated
			[219545] = true, -- Execute
		},
		cooldown = 26000, -- 26s
	},
	shieldThrow = 221945,

	-- Boss 3
	knot = {
		claim = 213410,
		carry = 213477,
		passage = {
			[222677] = true, -- Carrier
			[218492] = true, -- Bystanders
		},
		passageName = 222677,
	},
	links = {
		[223028] = 1,
		[223029] = 2,
	},
	flux = {
		start = 214570,
		carry = 214597,
		overloaded = 214745,
		interval = 62500, -- 62.5s
		initial = 16500, -- 16.5s
		maxCarry = 15000, -- 15s
	},
	tempest = {
		start = 215107,
		interval = 32000, -- 32s
		initial = 28000, -- 28s
	},
	xoryn = 219230,
	charge = 214542,
	chain = 219179,
	weakening = 222613,
}

local Vars = {
	mirrorSide = {
		light = false,
		dark = false,
	},
	nextJump = nil,
	knot = {
		unitId = nil,
		carryEnd = 0,
		duration = 0,
		passage = { },
	},
	flux = {
		state = nil,
		next = 0,
		unitId = nil,
		carryStart = 0,
		carryEnd = 0,
		overloaded = { },
	},
	link1 = "",
	platformHealth = nil,
	tempestNext = 0,
}

local function Identifier( x )
	return Module.ID .. x
end

local function GroupPanelPoll( )
	local currentTime = GetGameTimeMilliseconds()
	local passage = Vars.knot.passage
	local overloaded = Vars.flux.overloaded

	for i = 1, GetGroupSize() do
		local unitTag = GetGroupUnitTagByIndex(i)
		local unitId = LCA.IdentifyGroupUnitTag(unitTag)

		if (unitId) then
			local color, hasPassage, hasOverloaded
			local timers = { }

			if (passage[unitId]) then
				local remaining = passage[unitId] - currentTime
				if (remaining > 0) then
					hasPassage = true
					table.insert(timers, LCA.FormatTime(remaining, LCA.TIME_FORMAT_COMPACT))
				end
			end

			if (overloaded[unitId]) then
				local remaining = overloaded[unitId] - currentTime
				if (remaining > 0) then
					hasOverloaded = true
					table.insert(timers, LCA.FormatTime(remaining, LCA.TIME_FORMAT_COMPACT))
				end
			end

			if (hasPassage and hasOverloaded) then
				color = 0xFF00FF99
			elseif (hasPassage) then
				color = 0xFF000099
			elseif (hasOverloaded) then
				color = 0x0099FF99
			end

			CA2.GroupPanelUpdate(unitTag, unitId, color, table.concat(timers, "/"))
		end
	end
end

local function StatusPoll_B1( )
	local results = { }
	for i = 1, 2 do
		local current, _, effectiveMax = GetUnitPower("boss" .. i, COMBAT_MECHANIC_FLAGS_HEALTH)
		if (effectiveMax > 0) then
			table.insert(results, string.format("|c%06X%d%%|r", Data.boss1Colors[i], zo_floor(100 * current / effectiveMax)))
		end
	end
	CA2.StatusSetCellText(1, 2, table.concat(results, " / "))
end

local function StatusPoll_B2( )
	if (Vars.nextJump) then
		local remaining = Vars.nextJump - GetGameTimeMilliseconds()
		local ratio = LCA.Clamp(remaining / Data.jump.cooldown, 0, 1)
		CA2.StatusModifyCell(1, 2, {
			text = LCA.FormatTime(remaining, LCA.TIME_FORMAT_SHORT),
			color = LCA.PackRGBA(LCA.HSLToRGB(ratio / 3, 1, 0.5, 1)),
		})
	else
		CA2.StatusSetRowHidden(1, true)
	end
end

local function StatusInit_B3( )
	for i = 1, 3 do
		CA2.StatusModifyCell(i, 2, {
			alignment = TEXT_ALIGN_RIGHT,
			minWidth = "00s",
		})
		if (i < 3) then
			CA2.StatusSetCellText(i, 0, " ")
		end
		if (i > 1) then
			CA2.StatusSetRowHidden(i, true)
		end
	end
end

local function StatusPoll_B3( )
	local currentTime = GetGameTimeMilliseconds()

	if (Vars.knot.unitId) then
		local remaining = Vars.knot.carryEnd - currentTime
		local ratio = LCA.Clamp(remaining / Vars.knot.duration, 0, 1)
		local _, name = LCA.IdentifyGroupUnitId(Vars.knot.unitId, true)
		CA2.StatusModifyCell(1, 2, {
			text = LCA.FormatTime(remaining, LCA.TIME_FORMAT_SHORT),
			color = LCA.PackRGBA(LCA.HSLToRGB(ratio / 3, 1, 0.5, 1)),
		})
		CA2.StatusSetCellText(1, 3, name)
	else
		CA2.StatusModifyCell(1, 2, {
			text = LCA.GetLocalizedData(STRINGS.noCarrier),
			color = 0xFF0000FF,
		})
		CA2.StatusSetCellText(1, 3, "")
	end

	local passage = Vars.knot.passage
	if (passage.selfId and passage[passage.selfId]) then
		local cdRemaining = passage[passage.selfId] - currentTime
		if (cdRemaining > 0) then
			CA2.StatusModifyCell(1, 0, {
				text = string.format(LCA.GetLocalizedData(STRINGS.cooldownLabel), LCA.FormatTime(cdRemaining, LCA.TIME_FORMAT_LONG)),
				color = 0xCC0000FF,
			})
		else
			CA2.StatusSetCellText(1, 0, " ")
		end
	end

	if (Vars.flux.state) then
		if (Vars.flux.state == 1) then
			CA2.StatusSetCellText(2, 2, " ")
			CA2.StatusSetCellText(2, 3, "")
		elseif (Vars.flux.state == 2) then
			local carried = currentTime - Vars.flux.carryStart
			local ratio = LCA.Clamp(1 - carried / Data.flux.maxCarry, 0, 1)
			CA2.StatusSetCellText(2, 2, LCA.FormatTime(Vars.flux.carryEnd - currentTime, LCA.TIME_FORMAT_SHORT))
			CA2.StatusSetCellText(2, 3, string.format("%s [|c%06X%s|r]", select(2, LCA.IdentifyGroupUnitId(Vars.flux.unitId, true)), LCA.PackRGB(LCA.HSLToRGB(ratio / 3, 1, 0.5)), LCA.FormatTime(carried, LCA.TIME_FORMAT_SHORT)))
		elseif (Vars.flux.state == 3) then
			CA2.StatusSetCellText(2, 2, LCA.FormatTime(Vars.flux.next - currentTime, LCA.TIME_FORMAT_SHORT))
			CA2.StatusSetCellText(2, 3, LCA.GetLocalizedData(STRINGS.nextFlux))
		end

		local overloaded = Vars.flux.overloaded
		if (overloaded.selfId and overloaded[overloaded.selfId]) then
			local cdRemaining = overloaded[overloaded.selfId] - currentTime
			if (cdRemaining > 0) then
				CA2.StatusModifyCell(2, 0, {
					text = string.format(LCA.GetLocalizedData(STRINGS.cooldownLabel), LCA.FormatTime(cdRemaining, LCA.TIME_FORMAT_LONG)),
					color = 0xCC0000FF,
				})
			else
				CA2.StatusSetCellText(2, 0, " ")
			end
		end

		CA2.StatusSetCellText(3, 2, LCA.FormatTime(Vars.tempestNext - currentTime, LCA.TIME_FORMAT_SHORT))
		CA2.StatusSetCellText(3, 3, LCA.GetLocalizedData(STRINGS.nextTempest))
	else
		local shouldHide = true

		if (not Vars.platformHealth and GetUnitName("boss2") ~= "") then
			Vars.platformHealth = select(2, GetUnitPower("boss2", COMBAT_MECHANIC_FLAGS_HEALTH))
		end

		if (Vars.platformHealth) then
			local results = { }
			for i = 1, 4 do
				local current, max, effectiveMax = GetUnitPower("boss" .. i, COMBAT_MECHANIC_FLAGS_HEALTH)
				if (max == Vars.platformHealth) then
					table.insert(results, string.format("%d%%", zo_floor(100 * current / effectiveMax)))
				end
			end
			if (#results > 0) then
				CA2.StatusSetCellText(2, 2, table.concat(results, " / "))
				shouldHide = false
			end
		end

		CA2.StatusSetRowHidden(2, shouldHide)
	end

	if (CA2.GroupPanelIsEnabled()) then
		GroupPanelPoll()
	end
end

local function MirrorSide( _, result, _, _, _, _, _, _, _, _, _, _, _, _, _, _, abilityId )
	if (result == ACTION_RESULT_EFFECT_GAINED_DURATION) then
		if (abilityId == Data.mirrorSide.light) then
			Vars.mirrorSide.light = true
			Vars.mirrorSide.dark = false
		else
			Vars.mirrorSide.light = false
			Vars.mirrorSide.dark = true
		end
	elseif (result == ACTION_RESULT_EFFECT_FADED) then
		if (abilityId == Data.mirrorSide.light) then
			Vars.mirrorSide.light = false
		else
			Vars.mirrorSide.dark = false
		end
	end
end

local function HandleFluxStart( initial )
	if (not Vars.flux.state) then
		CA2.StatusSetCellText(2, 1, LCA.GetAbilityName(Data.flux.start))
		CA2.StatusSetRowHidden(3, false)
	end
	local currentTime = GetGameTimeMilliseconds()
	if (initial) then
		Vars.flux.state = 3
		Vars.flux.next = currentTime + Data.flux.initial
		Vars.tempestNext = currentTime + Data.tempest.initial
	else
		Vars.flux.state = 1
		Vars.flux.next = currentTime + Data.flux.interval
	end
end

function Module:PostLoad( )
	-- Events that need to be tracked at all times, even out of combat

	EVENT_MANAGER:RegisterForEvent(Identifier("B1_L"), EVENT_COMBAT_EVENT, MirrorSide)
	EVENT_MANAGER:AddFilterForEvent(Identifier("B1_L"), EVENT_COMBAT_EVENT, REGISTER_FILTER_ABILITY_ID, Data.mirrorSide.light)
	EVENT_MANAGER:AddFilterForEvent(Identifier("B1_L"), EVENT_COMBAT_EVENT, REGISTER_FILTER_TARGET_COMBAT_UNIT_TYPE, COMBAT_UNIT_TYPE_PLAYER)

	EVENT_MANAGER:RegisterForEvent(Identifier("B1_D"), EVENT_COMBAT_EVENT, MirrorSide)
	EVENT_MANAGER:AddFilterForEvent(Identifier("B1_D"), EVENT_COMBAT_EVENT, REGISTER_FILTER_ABILITY_ID, Data.mirrorSide.dark)
	EVENT_MANAGER:AddFilterForEvent(Identifier("B1_D"), EVENT_COMBAT_EVENT, REGISTER_FILTER_TARGET_COMBAT_UNIT_TYPE, COMBAT_UNIT_TYPE_PLAYER)

	EVENT_MANAGER:RegisterForEvent(Identifier("B3_START"), EVENT_COMBAT_EVENT, function( )
		local settings = self:GetSettings()
		if (settings.cooldownPanel and LCA.isVet) then
			CA2.GroupPanelEnable({
				headerText = zo_strformat("<<t:1>> / <<t:2>>", GetAbilityName(Data.knot.passageName), GetAbilityName(Data.flux.overloaded)),
				statWidth = 48,
				useRange = false,
			})
		end
		if (settings.passageProtection and LCA.isVet) then
			self:ToggleSplinteredPassageBlock(true)
		end
		self:StartListening()
	end)
	EVENT_MANAGER:AddFilterForEvent(Identifier("B3_START"), EVENT_COMBAT_EVENT, REGISTER_FILTER_COMBAT_RESULT, ACTION_RESULT_EFFECT_GAINED_DURATION)
	EVENT_MANAGER:AddFilterForEvent(Identifier("B3_START"), EVENT_COMBAT_EVENT, REGISTER_FILTER_ABILITY_ID, Data.knot.claim)
end

function Module:PreUnload( )
	EVENT_MANAGER:UnregisterForEvent(Identifier("B1_L"), EVENT_COMBAT_EVENT)
	EVENT_MANAGER:UnregisterForEvent(Identifier("B1_D"), EVENT_COMBAT_EVENT)
	EVENT_MANAGER:UnregisterForEvent(Identifier("B3_START"), EVENT_COMBAT_EVENT)
end

function Module:PreStartListening( )
	Vars.knot.passage = { }
	Vars.flux.overloaded = { }
	Vars.flux.state = nil
	Vars.platformHealth = nil
end

function Module:PostStopListening( )
	CA2.StatusDisable()
	CA2.GroupPanelDisable()
	self:ToggleSplinteredPassageBlock(false)
end

function Module:ProcessCombatEvents( result, isError, abilityName, abilityGraphic, abilityActionSlotType, sourceName, sourceType, targetName, targetType, hitValue, powerType, damageType, log, sourceUnitId, targetUnitId, abilityId, overflow )
	-- General
	if (result == ACTION_RESULT_EFFECT_GAINED_DURATION and targetType == COMBAT_UNIT_TYPE_PLAYER and abilityId == Data.refresh) then
		CA2.ScreenBorderEnable(0x99FFFFCC, 350, "u42refresh")
		LCA.PlaySounds("DUEL_FORFEIT")
	elseif (result == ACTION_RESULT_EFFECT_GAINED_DURATION and targetType == COMBAT_UNIT_TYPE_PLAYER and abilityId == Data.darkness) then
		CA1.Alert(nil, LCA.GetAbilityName(abilityId), 0xCC3333FF, SOUNDS.CHAMPION_POINTS_COMMITTED, 1500)

	-- Boss 1
	elseif (result == ACTION_RESULT_BEGIN and Data.boss1Abilities[abilityId]) then
		CA2.StatusEnable({
			ownerId = "u42b1",
			rowLabels = GetString("SI_ATTRIBUTES", ATTRIBUTE_HEALTH),
			pollingFunction = StatusPoll_B1,
		})
	elseif (result == ACTION_RESULT_BEGIN and Data.annihilation[abilityId] and hitValue <= 2000) then
		local side = Data.annihilation[abilityId]
		if (Vars.mirrorSide[side]) then
			CA1.Alert(nil, LCA.GetAbilityName(abilityId), 0x66CCFFFF, SOUNDS.CHAMPION_POINTS_COMMITTED, 3500)
		end
	elseif (result == ACTION_RESULT_EFFECT_GAINED_DURATION and targetType == COMBAT_UNIT_TYPE_PLAYER and Data.luster[abilityId] and not LCA.isTank and self:GetSettings().extraAlerts) then
		CA1.Alert(nil, LCA.GetAbilityName(abilityId), Data.luster[abilityId], SOUNDS.CHAMPION_POINTS_COMMITTED, 1500)
	elseif (result == ACTION_RESULT_EFFECT_GAINED and Data.summons[abilityId] and LCA.isTank) then
		local data = Data.summons[abilityId]
		if (Vars.mirrorSide[data.side]) then
			CA1.Alert(LCA.GetAbilityName(abilityId), zo_strformat(SI_INTERACT_PROMPT_FORMAT_UNIT_NAME, data.name), data.color, SOUNDS.CHAMPION_POINTS_COMMITTED, 2500)
		end
	elseif (result == ACTION_RESULT_BEGIN and Data.shear[abilityId] and targetType ~= COMBAT_UNIT_TYPE_PLAYER and not LCA.isTank and self:GetSettings().extraAlerts) then
		if (Vars.mirrorSide.light or CA2.StatusGetOwnerId() ~= "u42b1") then
			CA1.Alert(nil, LCA.GetAbilityName(Data.meteor), 0x00FFFFFF, SOUNDS.CHAMPION_POINTS_COMMITTED, 2000)
		end

	-- Boss 2
	elseif (result == ACTION_RESULT_BEGIN and abilityId == Data.fateSealer and (targetType == COMBAT_UNIT_TYPE_PLAYER or LCA.DoesPlayerHaveTauntSlotted())) then
		CA1.Alert(nil, LCA.GetAbilityName(abilityId), 0x66CCFFFF, SOUNDS.CHAMPION_POINTS_COMMITTED, 2500)
	elseif (result == ACTION_RESULT_BEGIN and abilityId == Data.jump.id) then
		local unitTag, name = LCA.IdentifyGroupUnitId(targetUnitId)
		if (unitTag) then
			Vars.nextJump = GetGameTimeMilliseconds() + Data.jump.cooldown
			if (targetType == COMBAT_UNIT_TYPE_PLAYER or LCA.GetDistance("player", unitTag) < 6) then
				CA1.Alert(LCA.GetAbilityName(abilityId), name, 0xCC3399FF, SOUNDS.CHAMPION_POINTS_COMMITTED, 2000)
			end
			CA2.StatusEnable({
				ownerId = "u42b2",
				rowLabels = LCA.GetAbilityName(abilityId),
				pollingFunction = StatusPoll_B2,
			})
		end
	elseif (result == ACTION_RESULT_BEGIN and Data.jump.escape[abilityId]) then
		-- Hide timer when Xoryn leaves
		Vars.nextJump = nil
	elseif (result == ACTION_RESULT_BEGIN and abilityId == Data.shieldThrow and not LCA.isTank and self:GetSettings().extraAlerts) then
		CA1.Alert(nil, LCA.GetAbilityName(abilityId), 0x00FFFFFF, nil, 2000)
		LCA.PlaySounds("DUEL_BOUNDARY_WARNING", 2, 150, "DUEL_BOUNDARY_WARNING", 3, 250, "FRIEND_INVITE_RECEIVED", 3)

	-- Boss 3
	elseif (abilityId == Data.knot.carry) then
		if (result == ACTION_RESULT_EFFECT_GAINED_DURATION) then
			Vars.knot.unitId = targetUnitId
			Vars.knot.carryEnd = GetGameTimeMilliseconds() + hitValue
			Vars.knot.duration = hitValue
			CA2.StatusEnable({
				ownerId = "u42b3",
				rowLabels = { LCA.GetAbilityName(Data.knot.carry), GetString("SI_ATTRIBUTES", ATTRIBUTE_HEALTH), LCA.GetAbilityName(Data.tempest.start) },
				pollingFunction = StatusPoll_B3,
				initFunction = StatusInit_B3,
			})
		elseif (result == ACTION_RESULT_EFFECT_FADED) then
			Vars.knot.unitId = nil
		end
	elseif (Data.knot.passage[abilityId]) then
		local passage = Vars.knot.passage
		if (result == ACTION_RESULT_EFFECT_GAINED_DURATION) then
			passage[targetUnitId] = GetGameTimeMilliseconds() + hitValue
		elseif (result == ACTION_RESULT_EFFECT_FADED) then
			passage[targetUnitId] = nil
		end
		if (targetType == COMBAT_UNIT_TYPE_PLAYER) then
			passage.selfId = targetUnitId
		end
	elseif (result == ACTION_RESULT_EFFECT_GAINED and Data.links[abilityId]) then
		local _, name = LCA.IdentifyGroupUnitId(targetUnitId, true)
		if (Data.links[abilityId] == 1) then
			Vars.link1 = name
		else
			CA1.Alert(LCA.GetAbilityName(abilityId), string.format("%s / %s", Vars.link1, name), 0xCC3399FF, SOUNDS.CHAMPION_POINTS_COMMITTED, 3500)
			Vars.link1 = ""
		end
	elseif (result == ACTION_RESULT_EFFECT_GAINED_DURATION and abilityId == Data.xoryn) then
		HandleFluxStart(true)
	elseif (result == ACTION_RESULT_BEGIN and abilityId == Data.flux.start) then
		local _, name = LCA.IdentifyGroupUnitId(targetUnitId, true)
		CA1.Alert(LCA.GetAbilityName(abilityId), name, 0x00CC99FF, SOUNDS.CHAMPION_POINTS_COMMITTED, 2000)
		HandleFluxStart()
	elseif (abilityId == Data.flux.carry) then
		if (result == ACTION_RESULT_EFFECT_GAINED_DURATION) then
			local currentTime = GetGameTimeMilliseconds()
			Vars.flux.state = 2
			Vars.flux.unitId = targetUnitId
			Vars.flux.carryStart = currentTime
			Vars.flux.carryEnd = currentTime + hitValue
		elseif (result == ACTION_RESULT_EFFECT_FADED and Vars.flux.unitId == targetUnitId) then
			Vars.flux.state = 3
			Vars.flux.unitId = nil
		end
	elseif (abilityId == Data.flux.overloaded) then
		local overloaded = Vars.flux.overloaded
		if (result == ACTION_RESULT_EFFECT_GAINED_DURATION) then
			overloaded[targetUnitId] = GetGameTimeMilliseconds() + hitValue
		elseif (result == ACTION_RESULT_EFFECT_FADED) then
			overloaded[targetUnitId] = nil
		end
		if (targetType == COMBAT_UNIT_TYPE_PLAYER) then
			overloaded.selfId = targetUnitId
		end
	elseif (result == ACTION_RESULT_BEGIN and abilityId == Data.tempest.start) then
		CA1.Alert(nil, LCA.GetAbilityName(abilityId), 0x66CCFFFF, SOUNDS.CHAMPION_POINTS_COMMITTED, 2000)
		Vars.tempestNext = GetGameTimeMilliseconds() + Data.tempest.interval
	elseif (result == ACTION_RESULT_BEGIN and abilityId == Data.charge and hitValue > 1000 and targetType ~= COMBAT_UNIT_TYPE_PLAYER and not LCA.isTank and self:GetSettings().extraAlerts) then
		zo_callLater(function()
			CA1.Alert(nil, LCA.GetAbilityName(Data.chain), 0x00FFFFFF, nil, 3250)
			LCA.PlaySounds("DUEL_BOUNDARY_WARNING", 1, 200, "DUEL_BOUNDARY_WARNING", 2, 200, "DUEL_BOUNDARY_WARNING", 3, 200, "DUEL_BOUNDARY_WARNING", 4)
		end, 1500)
	elseif (result == ACTION_RESULT_EFFECT_GAINED_DURATION and abilityId == Data.weakening and (targetType == COMBAT_UNIT_TYPE_PLAYER or LCA.DoesPlayerHaveTauntSlotted())) then
		local _, name = LCA.IdentifyGroupUnitId(targetUnitId, true)
		CA1.Alert(LCA.GetAbilityName(abilityId), name, 0xCC3399FF, SOUNDS.CHAMPION_POINTS_COMMITTED, 2000)
	end
end

do
	local action = LCA.GetLocalizedData(STRINGS["74865733-0-34170"])
	local name = LCA.GetLocalizedData(STRINGS["87370069-0-34170"])

	local shouldBlock = function( )
		local passage = Vars.knot.passage
		if (passage.selfId and passage[passage.selfId]) then
			return passage[passage.selfId] > GetGameTimeMilliseconds()
		end
	end

	function Module:ToggleSplinteredPassageBlock( enable )
		if (enable) then
			LCA.RegisterInteractionBlock(self.ID, action, name, shouldBlock)
		else
			LCA.RegisterInteractionBlock(self.ID)
		end
	end
end

function Module:GetSettingsControls( )
	local extraAlertsTooltipNames = { }
	for _, abilityId in ipairs(Data.extraAlertsTooltipIds) do
		table.insert(extraAlertsTooltipNames, LCA.GetAbilityName(abilityId))
	end

	local settings = self:GetSettings()

	return {
		--------------------
		{
			type = "checkbox",
			name = LCA.GetLocalizedData(STRINGS.extraAlerts),
			tooltip = table.concat(extraAlertsTooltipNames, "\n"),
			getFunc = function() return settings.extraAlerts end,
			setFunc = function(enabled) settings.extraAlerts = enabled end,
		},
		--------------------
		{
			type = "checkbox",
			name = LCA.GetLocalizedData(STRINGS.cooldownPanel),
			getFunc = function() return settings.cooldownPanel end,
			setFunc = function(enabled) settings.cooldownPanel = enabled end,
		},
		--------------------
		{
			type = "checkbox",
			name = LCA.GetLocalizedData(STRINGS.passageProtection),
			getFunc = function() return settings.passageProtection end,
			setFunc = function(enabled) settings.passageProtection = enabled end,
		},
	}
end

function Module:GetDebugData( )
	return Vars
end

CA2.RegisterModule(Module)
